from __future__ import absolute_import, unicode_literals

import logging

import pykka

from mopidy import backend
from mopidy.models import Album, Artist, Image, Ref, SearchResult, Track

from . import Extension
from .cdrom import CD_PROTOCOL, UNKNOWN_DISC, CdRom


logger = logging.getLogger(__name__)

ROOT_URI = 'cd:/'


class CdBackend(pykka.ThreadingActor, backend.Backend):

    uri_schemes = [Extension.ext_name]

    def __init__(self, config, audio):
        super(CdBackend, self).__init__()
        self.library = CdLibrary(backend=self)
        self.playback = CdPlayback(audio=audio, backend=self)


class CdLibrary(backend.LibraryProvider):

    root_directory = Ref.directory(uri=ROOT_URI, name='CD')
    cdrom = CdRom()

    def browse(self, uri):
        self.refresh()

        return [
            Ref.track(uri=ROOT_URI + str(track.number), name=track.title)
            for track in self.cdrom.disc.tracks
        ]

    def lookup(self, uri):
        disc = self.cdrom.disc
        album = CdLibrary._make_album(disc)
        track_path = uri.lstrip(ROOT_URI)
        if track_path:
            try:
                track_number = int(track_path)
            except ValueError as e:
                logger.warning('Invalid uri %s: %s', uri, e)
                return []
            else:
                logger.debug('CD track #%d selected', track_number)
                track = disc.tracks[track_number - 1]
                return [self._make_track(album, track)]
        else:
            logger.debug('All CD tracks selected')
            return [self._make_track(album, track) for track in disc.tracks]

    def get_images(self, uris):
        images = {Image(uri=img) for img in self.cdrom.disc.images}
        return {uri: images for uri in uris}

    def refresh(self, uri=None):
        self.cdrom.read()

    def search(self, query=None, uris=None, exact=False):
        def match(subvalue, value):
            if exact:
                return subvalue == value
            else:
                return subvalue.lower() in value.lower()

        if uris and all(uri not in ROOT_URI for uri in uris):
            return None

        self.refresh()

        disc = self.cdrom.disc
        if disc == UNKNOWN_DISC:
            return None

        any_query = query.get('any')
        album_query = query.get('album') or any_query or ()
        artist_query = query.get('artist') or any_query or ()
        track_name_query = query.get('track_name') or any_query or ()

        disc_album = CdLibrary._make_album(disc)
        return SearchResult(
            albums=[disc_album] if any(
                match(album, disc_album.name)
                for album in album_query
            ) else (),
            artists=[
                disc_artist
                for disc_artist in disc_album.artists
                if any(
                    match(artist, disc_artist.name)
                    for artist in artist_query
                )
            ],
            tracks=[
                CdLibrary._make_track(disc_album, track)
                for track in disc.tracks
                if any(
                    match(track_name, track.title)
                    for track_name in track_name_query
                )
            ]
        )

    @staticmethod
    def _make_album(disc):
        return Album(
            uri=ROOT_URI,
            musicbrainz_id=disc.id,
            name=disc.title,
            date=disc.year,
            artists={CdLibrary._make_artist(ar) for ar in disc.artists},
            images=disc.images,
            num_discs=disc.discs,
            num_tracks=len(disc.tracks)
        )

    @staticmethod
    def _make_artist(artist_tuple):
        return Artist(
            uri=ROOT_URI,
            musicbrainz_id=artist_tuple.id,
            name=artist_tuple.name,
            sortname=artist_tuple.sortname
        )

    @staticmethod
    def _make_track(album, track_tuple):
        return Track(
            uri=ROOT_URI + str(track_tuple.number),
            musicbrainz_id=track_tuple.id,
            name=track_tuple.title,
            length=track_tuple.duration,
            track_no=track_tuple.number,
            disc_no=track_tuple.disc_number,
            album=album,
            artists={CdLibrary._make_artist(ar) for ar in track_tuple.artists}
        )


class CdPlayback(backend.PlaybackProvider):

    def translate_uri(self, uri):
        return uri.replace(ROOT_URI, CD_PROTOCOL)



from __future__ import unicode_literals

import os

from mopidy import config, ext


__version__ = '0.5'


class Extension(ext.Extension):

    dist_name = 'Mopidy-Cd'
    ext_name = 'cd'
    version = __version__

    def get_default_config(self):
        conf_file = os.path.join(os.path.dirname(__file__), 'ext.conf')
        return config.read(conf_file)

    def setup(self, registry):
        from .backend import CdBackend
        registry.add('backend', CdBackend)



from __future__ import absolute_import, unicode_literals

import logging
import math

import musicbrainzngs

from collections import Mapping, namedtuple
from datetime import timedelta

from mopidy.audio.scan import Scanner
from mopidy.exceptions import ScannerError

from . import Extension


logger = logging.getLogger(__name__)

Artist = namedtuple('Artist', 'id name sortname')
Track = namedtuple('Track', 'id title number disc_number duration artists')
Disc = namedtuple('Disc', 'id discid title year discs artists images tracks')

CD_PROTOCOL = 'cdda://'
UNKNOWN_DISC = Disc(None, None, 'Unknown CD', '', 1, (), (), ())


musicbrainzngs.set_useragent(Extension.dist_name, Extension.version)


class DiscID(object):

    id = None
    toc = ''
    tracks = ()

    def __init__(self):
        try:
            # see GstAudioCdSrc docs for emitted tags
            tags = Scanner().scan(uri=CD_PROTOCOL).tags
            toc = [
                int(i, 16) for i in tags['musicbrainz-discid-full'][0].split()
            ]
            offsets = toc[3:] + toc[2:3]

            self.id = tags['musicbrainz-discid'][0]
            self.toc = ' '.join(str(i) for i in toc)
            self.tracks = [
                (i + 1, DiscID._to_seconds(offsets[i + 1] - offsets[i]))
                for i in range(len(offsets) - 1)
            ]
        except (LookupError, ScannerError) as e:
            logger.info('Error identifying disc: %s', e)

    @staticmethod
    def _to_seconds(sectors):
        return int(math.floor(sectors / 75.0 + 0.5))


class CdRom(object):

    disc = UNKNOWN_DISC

    def read(self):
        discid = DiscID()
        if discid.id:
            logger.debug(
                'Read disc: MusicBrainz DiscID %s, %d tracks',
                discid.id, len(discid.tracks)
            )
        else:
            self.disc = UNKNOWN_DISC
            return

        # use cached disc info if possible
        if self.disc.discid == discid.id:
            return

        try:
            mbrainz_info = musicbrainzngs.get_releases_by_discid(
                id=discid.id,
                toc=discid.toc,
                includes=['artist-credits', 'recordings'],
                cdstubs=False
            )
            # mbrainz_info is either
            # a {disc: {release-list: [...]}, ...} dict or
            # a {release-list: [...]} dict when matched by ToC
            release = mbrainz_info.get('disc', mbrainz_info)['release-list'][0]
            try:
                images = musicbrainzngs.get_image_list(release['id'])
            except musicbrainzngs.ResponseError as e:
                logger.debug('Error getting CD images from MusicBrainz: %s', e)
                images = {'images': ()}

            self.disc = Disc(
                id=release['id'],
                discid=discid.id,
                title=release['title'],
                discs=release['medium-count'],
                year=release['date'],
                images=CdRom._extract_images(images['images']),
                artists=CdRom._extract_artists(release['artist-credit']),
                tracks=CdRom._extract_tracks(discid, release['medium-list'])
            )
        except (LookupError, musicbrainzngs.WebServiceError) as e:
            logger.info('Error accessing MusicBrainz: %s', e)
            self.disc = UNKNOWN_DISC._replace(
                discid=discid.id,
                tracks=CdRom._extract_tracks(discid)
            )

    @staticmethod
    def _extract_artists(artist_credits):
        return {
            CdRom._make_artist(credit)
            for credit in artist_credits
            if isinstance(credit, Mapping)
        }

    @staticmethod
    def _extract_images(images_list):
        return {
            image['image']
            for image in images_list
            if image['front'] or image['back']
        }

    @staticmethod
    def _extract_tracks(discid, medium_list=()):
        def match_by_discid(medium):
            if medium['format'].lower() != 'cd':
                return False
            return any(disc['id'] == discid.id for disc in medium['disc-list'])

        cd = next(
            (medium for medium in medium_list if match_by_discid(medium)),
            None
        )
        if cd:
            disc_num = int(cd['position'])
            tracks = cd['track-list']
            return [CdRom._make_track_mbrainz(disc_num, tr) for tr in tracks]
        else:
            return [CdRom._make_track_discid(track) for track in discid.tracks]

    @staticmethod
    def _make_artist(artist_dict):
        artist = artist_dict['artist']
        return Artist(
            id=artist['id'],
            name=artist['name'],
            sortname=artist['sort-name']
        )

    @staticmethod
    def _make_track_mbrainz(disc_number, track):
        recording = track['recording']
        return Track(
            id=track['id'],
            title=recording['title'],
            number=int(track['number']),
            disc_number=disc_number,
            duration=int(track['length']),
            artists=CdRom._extract_artists(recording['artist-credit'])
        )

    @staticmethod
    def _make_track_discid(track):
        return Track(
            id=None,
            title='CD Track %d (%s)' % (track[0], timedelta(seconds=track[1])),
            number=track[0],
            disc_number=1,
            duration=track[1] * 1000,
            artists=()
        )
